/*
 * Copyright (c) 1982, 1986, 1988 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 *
 *	@(#)ts.c	7.5 (Berkeley) 7/9/88
 *
 * Hacked mercilessly by Brad Parker <brad@heeltoe.com> to run on the 730
 */

/*
 * TS11 tape driver
 */

#include "sys/param.h"

#include "../include/pte.h"

#include "tsreg.h"
#include "../dev/qbus/ubareg.h"

#include "lib/libsa/stand.h"
#include "lib/libkern/libkern.h"

#include "vaxstand.h"

extern int vax_boardtype;

#define	MAXCTLR		1		/* all addresses must be specified */
u_short	tsstd[MAXCTLR] = { (u_short)0772520 };

u_short	ts_uba;			/* Unibus address of ts structure */

struct ts {
	struct ts_cmd ts_cmd;
	struct ts_char ts_char;
	struct ts_sts ts_sts;
} ts;

static struct ts *ts_ubaddr;
static struct tsdevice *tsaddr;
static int ts_adapt;

tsopen(struct open_file *f, int adapt, int ctlr, int unit, int part)
{
	long i = 0;

//	if (adapt >= nuba)
//		return (EADAPT);
	if (ctlr >= MAXCTLR)
		return (ECTLR);
	/* TS11 only supports one transport per formatter */
	if (unit)
		return(EUNIT);

	ts_adapt = adapt;
	tsaddr = (struct tsdevice *)((char *)csrbase + (tsstd[ctlr] & 017777));

//#ifdef DEV_DEBUG
	printf("ts: tsaddr %x, csrbase %x\n", tsaddr, csrbase);
//#endif

#define DELAY(n) do { volatile int i; for (i=0;i<100*n;i++); } while(0);

	tsaddr->tssr = 0;
	while ((tsaddr->tssr & TS_SSR)==0) {
		DELAY(10);
		if (++i > 1000000) {
			printf("ts: not ready\n");
			return (ENXIO);
		}
	}
	if (tsaddr->tssr&TS_OFL) {
		printf("ts: offline\n");
		return (ENXIO);
	}
	if (tsaddr->tssr&TS_NBA) {
		int i;
		int *map;

#if 0
		if (ts_ubaddr == 0)
			ts_ubaddr = (struct ts *)ubasetup(ts_adapt, &ts, sizeof(ts), 2);
#else
		if (ts_ubaddr == 0) {
			/* arbitrarily use regs 490-492 */
			/* (de uses 0-n, ra uses 494-495) */
			map = (int *)nexaddr + 512;
			map[490] = PG_V | (((u_int)&ts) >> 9);
			ts_ubaddr = (struct ts *)((char *)(490 << 9) + 
						  (((u_int)&ts)&0x1ff));
		}
#endif
		ts_uba = (u_short)((long)ts_ubaddr + (((long)ts_ubaddr>>16)&03));
		ts.ts_char.char_addr = (int)&ts_ubaddr->ts_sts;
		ts.ts_char.char_size = sizeof(ts.ts_sts);
		ts.ts_char.char_mode = TS_ESS;
		ts.ts_cmd.c_cmd = TS_ACK|TS_SETCHR;
		i = (int)&ts_ubaddr->ts_char;
		ts.ts_cmd.c_loba = i;
		ts.ts_cmd.c_hiba = (i>>16)&3;
		ts.ts_cmd.c_size = sizeof(ts.ts_char);
		tsaddr->tsdb = ts_uba;
	}
	tsstrategy(f, TS_REW, 0, 0, 0, 0);
	if (part)
		tsstrategy(f, TS_SFORWF, 0, part, 0, 0);
	return (0);
}

tsclose(struct open_file *f)
{
	if (f->f_flags & F_WRITE) {
		printf("writing tape mark\n");
		tsstrategy(f, TS_WEOF, 0, 1, 0, 0);/* Write Tape Mark */
		tsstrategy(f, TS_WEOF, 0, 1, 0, 0);/* Write Tape Mark */
		tsstrategy(f, TS_SREVF, 0, 1, 0, 0);/* Skip Tape Marks Reverse */
	} else {
		/* should we rewind? */
//		tsstrategy(f, TS_REW, 0, 0, 0, 0);
	}
}

static char *
tssr_bits_text(unsigned bits)
{
	static char b[80];
	b[0] = 0;
	if (bits & (1 << 15)) strcat(b, "SC ");
	if (bits & (1 << 14)) strcat(b, "UPE ");
	if (bits & (1 << 13)) strcat(b, "SPE ");
	if (bits & (1 << 12)) strcat(b, "RMR ");
	if (bits & (1 << 11)) strcat(b, "NXM ");
	if (bits & (1 << 10)) strcat(b, "NBA ");
	if (bits & (1 <<  9)) strcat(b, "A17 ");
	if (bits & (1 <<  8)) strcat(b, "A16 ");
	if (bits & (1 <<  7)) strcat(b, "SSR ");
	if (bits & (1 <<  6)) strcat(b, "OFL ");
	if (bits & (1 <<  5)) strcat(b, "FC1 ");
	if (bits & (1 <<  4)) strcat(b, "FC0 ");
	if (bits & (1 <<  3)) strcat(b, "TC2 ");
	if (bits & (1 <<  2)) strcat(b, "TC1 ");
	if (bits & (1 <<  1)) strcat(b, "TC0 ");
	if (bits & (1 <<  0)) strcat(b, "-");
	return b;
}

int
tsstrategy(void *f, int func, daddr_t dblk,
    size_t size, void *buf, size_t *rsize)
{
	register int errcnt, info;

	//printf("tsstrategy; func %o, dblk %d, size %d, buf %x\n",
	//       func, (unsigned int)dblk, size, buf);

	errcnt = info = 0;
retry:
	while ((tsaddr->tssr & TS_SSR) == 0)
		DELAY(100);
	if (func == TS_REW || func == TS_SFORWF ||
	    func == TS_SREVF || func == TS_WEOF)
	{
		ts.ts_cmd.c_repcnt = size;
	} else {
#if 0
		info = ubasetup(ts_adapt, buf, size, 1);
#else
		int i, *map, npgs, mdata;

		/* map buffer using registers 400-4xx */
		/* (copies are 10k, so 400-419) */
		map = (int *)nexaddr + 512;
		npgs = (size >> VAX_PGSHIFT) + 1;
		mdata = (int)buf >> VAX_PGSHIFT;

		for (i = 0; i < npgs; i++)
			map[400+i] = PG_V | (mdata + i);

		info = (400 << 9) + (((u_int)buf)&0x1ff);
#endif
		ts.ts_cmd.c_size = size;
		ts.ts_cmd.c_loba = info;
		ts.ts_cmd.c_hiba = (info>>16)&3;
		if (func == F_READ)
			func = TS_RCOM;
		else if (func == F_WRITE)
			func = TS_WCOM;
	}
	ts.ts_cmd.c_cmd = TS_ACK|TS_CVC|func;
	tsaddr->tsdb = ts_uba;
	do
		DELAY(100)
	while ((tsaddr->tssr & TS_SSR) == 0);
#if 0
	if (info)
		ubafree(ts_adapt, info);
#endif
	if (ts.ts_sts.s_xs0 & TS_TMK) {
		*rsize = 0;
		return (0);
	}
	if (tsaddr->tssr & TS_SC) {
		printf("ts tape error: tssr=%x (%s), xs0=%x\n",
		       tsaddr->tssr, tssr_bits_text(tsaddr->tssr),
		       ts.ts_sts.s_xs0);
		if (errcnt++ == 10) {
			printf("ts: unrecovered error\n");
			return (-1);
		}
		if (func == TS_RCOM || func == TS_WCOM)
			func |= TS_RETRY;
		goto retry;
	}
	if (errcnt)
		printf("ts: recovered by retry\n");
	*rsize = size - ts.ts_sts.s_rbpcr;
	return (0);
}

